#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 29, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : metrics.py
project  : nadl
license  : GPL-3.0+

Some simple metrics.  For complex metrics, use sklearn.
"""
from functools import partial

import jax
import jax.numpy as jnp


@partial(jax.jit, static_argnames=("normalize",))
def accuracy(y_true: jax.Array, y_pred: jax.Array, normalize: bool = True) -> jax.Array:
  """Accuracy."""
  return jnp.mean(y_true == y_pred) if normalize else jnp.sum(y_true == y_pred)


def roc_curve(
  y_true: jax.Array, y_score: jax.Array
) -> tuple[jax.Array, jax.Array, jax.Array]:
  """Compute Receiver operating characteristic (ROC).  Cannot be jitted."""
  y_true = jnp.asarray(y_true)
  y_score = jnp.asarray(y_score)

  descending_order = jnp.argsort(y_score)[::-1]
  y_score = y_score[descending_order]
  y_true = y_true[descending_order]

  # TPR and FPR
  distinct_value_indices = jnp.argwhere(jnp.diff(y_score))
  threshold_idxs = jnp.r_[distinct_value_indices, y_true.size - 1]

  tps = jnp.cumsum(y_true)[threshold_idxs]
  fps = 1 + threshold_idxs - tps

  tpr = tps / jnp.sum(y_true)
  fpr = fps / jnp.sum(1 - y_true)

  thresholds = jnp.r_[jnp.inf, y_score[threshold_idxs]]

  tpr = jnp.r_[0, tpr]
  fpr = jnp.r_[0, fpr]

  return fpr, tpr, thresholds


def precision_recall_curve(
  y_true: jax.Array, y_score: jax.Array
) -> tuple[jax.Array, jax.Array, jax.Array]:
  """Compute precision-recall pairs for different probability thresholds."""
  y_true = jnp.asarray(y_true)
  y_score = jnp.asarray(y_score)

  descending_order = jnp.argsort(y_score)[::-1]
  y_score = y_score[descending_order]
  y_true = y_true[descending_order]

  # Precision and recall
  distinct_value_indices = jnp.argwhere(jnp.diff(y_score))
  threshold_idxs = jnp.r_[distinct_value_indices, y_true.size - 1]

  tps = jnp.cumsum(y_true)[threshold_idxs]
  fps = 1 + threshold_idxs - tps

  precision = tps / (tps + fps)
  recall = tps / jnp.sum(y_true)

  thresholds = jnp.r_[jnp.inf, y_score[threshold_idxs]]

  precision = jnp.r_[1, precision]
  recall = jnp.r_[0, recall]

  return precision, recall, thresholds


@jax.jit
def auc(fpr: jax.Array, tpr: jax.Array) -> jax.Array:
  """Compute Area Under the Curve (AUC) from the ROC curve."""
  # Sort by false positive rates
  sorted_indices = jnp.argsort(fpr)
  fpr, tpr = fpr[sorted_indices], tpr[sorted_indices]

  # Compute the area using the trapezoidal rule
  return jax.scipy.integrate.trapezoid(tpr, fpr)


@partial(jax.jit, static_argnames=("num_classes",))
def confusion_matrix(
  y_true: jax.Array, y_pred: jax.Array, num_classes: int = 2
) -> jax.Array:
  """Compute confusion matrix."""
  return (
    jnp.zeros((num_classes, num_classes), dtype=jnp.int32).at[(y_true, y_pred)].add(1)
  )


@partial(jax.jit, static_argnames=("num_classes", "eps"))
def precision_recall_fscore_support(
  y_true: jax.Array, y_pred: jax.Array, num_classes: int = 2, eps: float = 1e-8
) -> tuple[jax.Array, jax.Array, jax.Array, jax.Array]:
  """Compute precision, recall, F-score and support for each class."""
  # Create a confusion matrix
  conf_matrix = confusion_matrix(y_true, y_pred, num_classes)

  # Compute precision, recall for each class
  true_positives = jnp.diag(conf_matrix)
  predicted_positives = jnp.sum(conf_matrix, axis=0)
  actual_positives = jnp.sum(conf_matrix, axis=1)

  precision = true_positives / (predicted_positives + eps)
  recall = true_positives / (actual_positives + eps)

  # Compute F1 score
  f1_score = 2 * (precision * recall) / (precision + recall + eps)

  # Compute support for each class
  support = actual_positives

  return precision, recall, f1_score, support


@partial(jax.jit, static_argnames=("num_classes", "eps"))
def precision_score(
  y_true: jax.Array, y_pred: jax.Array, num_classes: int = 2, eps: float = 1e-8
) -> jax.Array:
  """Compute the precision."""
  precision, _, _, _ = precision_recall_fscore_support(y_true, y_pred, num_classes, eps)
  return precision


@partial(jax.jit, static_argnames=("num_classes", "eps"))
def recall_scroe(
  y_true: jax.Array, y_pred: jax.Array, num_classes: int = 2, eps: float = 1e-8
) -> jax.Array:
  """Compute the recall."""
  _, recall, _, _ = precision_recall_fscore_support(y_true, y_pred, num_classes, eps)
  return recall


@partial(jax.jit, static_argnames=("num_classes", "eps"))
def f1_score(
  y_true: jax.Array, y_pred: jax.Array, num_classes: int = 2, eps: float = 1e-8
) -> jax.Array:
  """Compute the F1 score."""
  _, _, f1_score, _ = precision_recall_fscore_support(y_true, y_pred, num_classes, eps)
  return f1_score


@partial(jax.jit, static_argnames=("num_classes", "eps"))
def support(
  y_true: jax.Array, y_pred: jax.Array, num_classes: int = 2, eps: float = 1e-8
) -> jax.Array:
  """Compute the support."""
  _, _, _, support = precision_recall_fscore_support(y_true, y_pred, num_classes, eps)
  return support


def roc_auc_score(y_true: jax.Array, y_score: jax.Array) -> jax.Array:
  """Compute Area Under the ROC of AUC from prediction scores."""
  # Compute ROC curve
  fpr, tpr, _ = roc_curve(y_true, y_score)
  return auc(fpr, tpr)


def average_precision_score(y_true: jax.Array, y_score: jax.Array) -> jax.Array:
  """Compute average precision from prediction scores."""
  # Compute precision-recall curve
  precision, recall, _ = precision_recall_curve(y_true, y_score)
  return jax.scipy.integrate.trapezoid(precision, recall)


def dice_coef(y_true: jax.Array, y_pred: jax.Array, eps: float = 1e-8) -> jax.Array:
  """Compute dice coefficient."""
  y_true = jnp.asarray(y_true)
  y_pred = jnp.asarray(y_pred)

  intersection = jnp.sum(y_true * y_pred)
  union = jnp.sum(y_true) + jnp.sum(y_pred)

  return (2.0 * intersection) / (union + eps)


def iou(y_true: jax.Array, y_pred: jax.Array, eps: float = 1e-8) -> jax.Array:
  """Compute intersection over union."""
  y_true = jnp.asarray(y_true)
  y_pred = jnp.asarray(y_pred)

  intersection = jnp.sum(y_true * y_pred)
  union = jnp.sum(y_true) + jnp.sum(y_pred) - intersection

  return intersection / (union + eps)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Dec  8, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : blocks.py
project  : nadl
license  : GPL-3.0+

Layer Blocks
"""
from collections.abc import Sequence
from functools import partial

import jax
import jax.numpy as jnp
from flax import linen

from typing import Literal


class ConvBlock(linen.Module):
  """Define a convolutional block with optional depthwise conv and residual conn."""

  features: int
  kernel_size: int = 3
  use_depthwise: bool = True

  @linen.compact
  def __call__(self, x: jax.Array, train: bool = False) -> jax.Array:
    """Apply the convolution block to the input."""
    ks = (self.kernel_size, self.kernel_size)
    if self.use_depthwise:
      x = linen.Conv(
        features=x.shape[-1],
        kernel_size=ks,
        padding="SAME",
        feature_group_count=x.shape[-1],
      )(x)
      x = linen.Conv(features=self.features, kernel_size=(1, 1))(x)
    else:
      x = linen.Conv(self.features, ks, padding="SAME")(x)

    x = linen.BatchNorm()(x, use_running_average=not train)
    return linen.PReLU()(x)


class DownSample(linen.Module):
  """Downsample block for UNet encoder."""

  features: int
  kernel_size: int = 3
  use_depthwise: bool = False
  pool: Literal["before", "after", None] = "after"
  pool_type: Literal["max", "avg"] = "max"

  @linen.compact
  def __call__(self, x: jax.Array, train: bool = False) -> jax.Array:
    """Apply downsampling to the input."""
    match self.pool_type:
      case "max":
        pool_fn = linen.max_pool
      case "avg":
        pool_fn = linen.avg_pool
      case _:
        raise ValueError(f"Unknown pool type: {self.pool_type}")
    pool_fn = partial(pool_fn, window_shape=(2, 2), strides=(2, 2), padding="VALID")
    if self.pool == "before":
      x = pool_fn(x)

    x = ConvBlock(
      features=self.features,
      kernel_size=self.kernel_size,
      use_depthwise=self.use_depthwise,
    )(x, train=train)

    if self.pool == "after":
      x = pool_fn(x)
    return x


class UpSample(linen.Module):
  """Upsample block for UNet decoder."""

  features: int
  kernel_size: int = 3
  use_depthwise: bool = False

  @linen.compact
  def __call__(self, x: jax.Array, skip: jax.Array, train: bool = False) -> jax.Array:
    """Apply upsampling to the input and concatenate with skip connection."""
    x = linen.ConvTranspose(
      features=self.features,
      kernel_size=(2, 2),
      strides=(2, 2),
    )(x)

    if x.shape != skip.shape:
      diff = skip.shape[1] - x.shape[1]
      x = jnp.pad(
        x,
        [
          (0, 0),
          (diff // 2, diff - diff // 2),
          (diff // 2, diff - diff // 2),
          (0, 0),
        ],
      )

    x = jnp.concatenate([x, skip], axis=-1)
    return ConvBlock(
      features=self.features,
      kernel_size=self.kernel_size,
      use_depthwise=self.use_depthwise,
    )(x, train=train)


class BasicBlock(linen.Module):
  """Define a basic residual block."""

  features: int
  strides: tuple = (1, 1)
  use_projection: bool = False

  @linen.compact
  def __call__(self, x: jax.Array, train: bool = False) -> jax.Array:
    """Apply the basic residual block to the input."""
    identity = x

    # First convolutional layer
    y = linen.Conv(
      self.features, kernel_size=(3, 3), strides=self.strides, padding="SAME"
    )(x)
    y = linen.BatchNorm()(y, use_running_average=not train)
    y = linen.relu(y)

    # Second convolutional layer
    y = linen.Conv(self.features, kernel_size=(3, 3), padding="SAME")(y)
    y = linen.BatchNorm()(y, use_running_average=not train)

    # Projection shortcut, if required
    if self.use_projection:
      identity = linen.Conv(
        self.features, kernel_size=(1, 1), strides=self.strides, padding="SAME"
      )(x)
      identity = linen.BatchNorm()(identity, use_running_average=not train)

    # Adding the shortcut
    y += identity
    return linen.relu(y)


class BottleneckBlock(linen.Module):
  """Define a bottleneck residual block."""

  features: int
  expansion: int = 4
  strides: tuple = (1, 1)
  use_projection: bool = False

  @linen.compact
  def __call__(self, x: jax.Array, train: bool = False) -> jax.Array:
    """Apply the bottleneck residual block to the input."""
    identity = x

    # First layer
    y = linen.Conv(self.features, kernel_size=(1, 1))(x)
    y = linen.BatchNorm()(y, use_running_average=not train)
    y = linen.relu(y)

    # Second layer
    y = linen.Conv(
      self.features, kernel_size=(3, 3), strides=self.strides, padding="SAME"
    )(y)
    y = linen.BatchNorm()(y, use_running_average=not train)
    y = linen.relu(y)

    # Third layer
    y = linen.Conv(self.features * self.expansion, kernel_size=(1, 1))(y)
    y = linen.BatchNorm()(y, use_running_average=not train)

    # Projection shortcut
    if self.use_projection:
      identity = linen.Conv(
        self.features * self.expansion, kernel_size=(1, 1), strides=self.strides
      )(x)
      identity = linen.BatchNorm()(identity, use_running_average=not train)

    y += identity
    return linen.relu(y)


def make_layer(
  block: type[linen.Module],
  features: int,
  num_blocks: int,
  strides: tuple[int, int],
  use_projection: bool = False,
) -> list[linen.Module]:
  """Make a layer with the given block."""
  return list(
    map(
      lambda i: block(
        features,
        strides=strides if i == 0 else (1, 1),
        use_projection=use_projection if i == 0 else False,
      ),
      range(num_blocks),
    )
  )


class ResNetBasic(linen.Module):
  """Define a ResNet w/ Basic Block."""

  out_channels: int
  num_blocks: Sequence[int] = (2, 2, 2, 2)  # 2,2,2,2 -> ResNet 18
  init_features: int = 64

  @linen.compact
  def __call__(self, x: jax.Array, train: bool = False) -> tuple[jax.Array, jax.Array]:
    """Apply the ResNet18 to the input."""
    x = linen.Conv(
      self.init_features, kernel_size=(7, 7), strides=(2, 2), padding="SAME"
    )(x)
    x = linen.BatchNorm()(x, use_running_average=not train)
    x = linen.relu(x)
    x = linen.max_pool(x, window_shape=(3, 3), strides=(2, 2), padding="SAME")

    for i, bn in enumerate(self.num_blocks):
      x = linen.Sequential(
        make_layer(
          BasicBlock, self.init_features * (2**i), bn, (2, 2), use_projection=i > 0
        )
      )(x, train=train)

    res = x
    x = linen.avg_pool(x, window_shape=(x.shape[1], x.shape[2]))
    x = linen.Dense(self.out_channels)(x)
    return x, res


class ResNetBottleneck(linen.Module):
  """Define a ResNet w/ Bottleneck Block."""

  out_channels: int
  num_blocks: Sequence[int] = (3, 4, 6, 3)  # 3,4,6,3 -> ResNet 50
  init_features: int = 64

  @linen.compact
  def __call__(self, x: jax.Array, train: bool = False) -> tuple[jax.Array, jax.Array]:
    """Apply the ResNet50 to the input."""
    x = linen.Conv(
      self.init_features, kernel_size=(7, 7), strides=(2, 2), padding="SAME"
    )(x)
    x = linen.BatchNorm()(x, use_running_average=not train)
    x = linen.relu(x)
    x = linen.max_pool(x, window_shape=(3, 3), strides=(2, 2), padding="SAME")

    for i, bn in enumerate(self.num_blocks):
      x = linen.Sequential(
        make_layer(
          BottleneckBlock, self.init_features * (2**i), bn, (2, 2), use_projection=i > 0
        )
      )(x, train=train)

    res = x
    x = linen.avg_pool(x, window_shape=(x.shape[1], x.shape[2]))
    x = linen.Dense(self.out_channels)(x)
    return x, res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 30, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : loops.py
project  : nadl
license  : GPL-3.0+

Train Eval Loops
"""
from typing import NamedTuple
from collections.abc import Hashable

from rich.console import Console
from rich.progress import (
  Progress,
  TaskID,
  TextColumn,
  TimeElapsedColumn,
  TimeRemainingColumn,
  BarColumn,
  ProgressColumn,
)
from rich.theme import Theme


DEF_LIGHT_THEME = Theme(
  {
    "bar.back": "#50616D",
    "bar.complete": "#789262",
    "bar.finished": "#057748",
  }
)


class PG(NamedTuple):
  """Progress."""

  pg: Progress
  console: Console
  tasks: dict[Hashable, TaskID]


def init_progress(
  pg: Progress | None = None,
  console: Console | None = None,
  total: bool = True,
  bar_width: int | None = 20,
  extra_columns: tuple[ProgressColumn, ...] = (),
  show_progress: bool = True,
  theme: Theme | None = None,
) -> PG:
  """Init progress bar."""
  if console is None:
    console = Console(theme=theme or DEF_LIGHT_THEME)
  if pg is None:
    pg = Progress(
      TextColumn(
        "{task.description}" + " - {task.completed}/{task.total}" if total else ""
      ),
      TimeRemainingColumn(),
      TimeElapsedColumn(),
      BarColumn(bar_width),
      console=console,
      disable=not show_progress,
    )
  pg.columns = pg.columns + extra_columns
  return PG(pg, pg.console, {})


def add_columns(pg: PG, columns: tuple[ProgressColumn, ...]) -> PG:
  """Add columns."""
  pg.pg.columns = pg.pg.columns + columns
  return pg


def test() -> None:
  """Test progress."""
  import time
  pg = init_progress()
  with pg.pg:
    t0 = pg.tasks[0] = pg.pg.add_task("Task 0", total=100)
    for i in range(100):
      pg.pg.advance(t0)
      time.sleep(0.5)



if __name__ == "__main__":
  test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 29, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : __init__.py
project  : nadl
license  : GPL-3.0+

NADL
"""

from .keys import (
  Keys,
  from_int_or_key,
  from_state,
  new_key,
  next_key,
  reverse,
  take,
)
from .loops import PG, init_progress
from .losses import dice_loss, softmax_focal_loss, sigmoid_focal_loss
from .metrics import (
  accuracy,
  auc,
  average_precision_score,
  confusion_matrix,
  dice_coef,
  f1_score,
  iou,
  precision_recall_curve,
  precision_recall_fscore_support,
  precision_score,
  recall_scroe,
  roc_auc_score,
  roc_curve,
  support,
)
from .preprocessing import (
  SCALER,
  identity_scaler,
  min_max_scaler,
  standard_scaler,
  normalizer,
  select_scaler,
)
from .utils import (
  classit,
  rle,
  rle_array,
)
from .blocks import (
  BasicBlock,
  BottleneckBlock,
  ConvBlock,
  DownSample,
  ResNetBasic,
  ResNetBottleneck,
  UpSample,
  make_layer,
)

__version__ = "0.4.1"

__all__ = [
  "BasicBlock",
  "BottleneckBlock",
  "ConvBlock",
  "DownSample",
  "Keys",
  "PG",
  "ResNetBasic",
  "ResNetBottleneck",
  "SCALER",
  "UpSample",
  "accuracy",
  "auc",
  "average_precision_score",
  "classit",
  "confusion_matrix",
  "dice_coef",
  "dice_loss",
  "f1_score",
  "from_int_or_key",
  "from_state",
  "identity_scaler",
  "init_progress",
  "iou",
  "make_layer",
  "min_max_scaler",
  "new_key",
  "next_key",
  "normalizer",
  "precision_recall_curve",
  "precision_recall_fscore_support",
  "precision_score",
  "recall_scroe",
  "reverse",
  "rle",
  "rle_array",
  "roc_auc_score",
  "roc_curve",
  "select_scaler",
  "sigmoid_focal_loss",
  "softmax_focal_loss",
  "standard_scaler",
  "support",
  "take",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 29, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : keys.py
project  : nadl
license  : GPL-3.0+

Keys.
"""
from collections import deque
from collections.abc import Sequence

import jax

from typing import NamedTuple


class Keys(NamedTuple):
  """JAX random key sequence."""

  key: jax.Array
  subkeys: deque[jax.Array]


def from_int_or_key(key: jax.Array | int) -> Keys:
  """Convert int or key to Keys."""
  if isinstance(key, int):
    return Keys(jax.random.key(key), deque())
  return Keys(key, deque())


new_key = from_int_or_key


def from_state(key: jax.Array, subkeys: Sequence[jax.Array]) -> Keys:
  """Convert state to Keys."""
  return Keys(key, deque(subkeys))


def reverse(keys: Keys, num: int) -> Keys:
  """Reverse the keys."""
  key, subkeys = keys
  while (num := num - 1) >= 0:
    if subkeys:
      subkeys.append(jax.random.fold_in(subkeys[-1], 0))
    else:
      subkeys.append(jax.random.fold_in(key, 0))
  return keys._replace(subkeys=subkeys)


def next_key(keys: Keys) -> Keys:
  """Get next key."""
  if not keys.subkeys:
    keys = reverse(keys, 1)
  return from_state(keys.subkeys.popleft(), keys.subkeys)


def take(keys: Keys, num: int) -> tuple[Keys, ...]:
  """Take num keys."""
  keys = reverse(keys, max(num - len(keys.subkeys), 0))
  return tuple(map(lambda _: next_key(keys), range(num)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Dec  6, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : preprocessing.py
project  : nadl
license  : GPL-3.0+

Preprocessing
"""
from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp

from typing import Literal, TypeAlias

SCALER: TypeAlias = Callable[[jax.Array], jax.Array]


def identity_scaler(arr: jax.Array, axis: int = 0) -> SCALER:
  """Get identity scaler."""
  del arr, axis
  def scaler(x: jax.Array) -> jax.Array:
    """Scaler."""
    return x

  return scaler


def min_max_scaler(arr: jax.Array, axis: int = 0) -> SCALER:
  """Get min max scaler."""
  min_, max_ = arr.min(axis=axis, keepdims=True), arr.max(axis=axis, keepdims=True)

  def scaler(x: jax.Array) -> jax.Array:
    """Scaler."""
    return (x - min_) / (max_ - min_)

  return scaler


def standard_scaler(arr: jax.Array, axis: int = 0) -> SCALER:
  """Get standard scaler."""
  mean, std = arr.mean(axis=axis, keepdims=True), arr.std(axis=axis, keepdims=True)

  def scaler(x: jax.Array) -> jax.Array:
    """Scaler."""
    return (x - mean) / std

  return scaler


def normalizer(
  arr: jax.Array, axis: int = 0, norm: Literal["l1", "l2", "max"] = "l2"
) -> SCALER:
  """Get normalizer."""
  match norm:
    case "l2":
      norm_value = jnp.sqrt(jnp.sum(jnp.square(arr), axis=axis, keepdims=True))
    case "l1":
      norm_value = jnp.sum(jnp.abs(arr), axis=axis, keepdims=True)
    case "max":
      norm_value = jnp.max(jnp.abs(arr), axis=axis, keepdims=True)
    case _:
      raise ValueError("norm should be 'l1', 'l2', or 'max'")

  def scaler(x: jax.Array) -> jax.Array:
    """Scaler."""
    max_val = jnp.maximum(norm_value, jnp.finfo(x.dtype).tiny)  # Avoid division by zero
    return x / max_val

  return scaler


def select_scaler(
  method: Literal["id", "minmax", "std", "l2_norm", "l1_norm", "max_norm"] = "minmax",
  axis: int = 0,
) -> Callable[[jax.Array], SCALER]:
  """Get scaler function."""
  match method:
    case "id":
      return partial(identity_scaler, axis=axis)
    case "minmax":
      return partial(min_max_scaler, axis=axis)
    case "std":
      return partial(standard_scaler, axis=axis)
    case "l2_norm":
      return partial(normalizer, norm="l2", axis=axis)
    case "l1_norm":
      return partial(normalizer, norm="l1", axis=axis)
    case "max_norm":
      return partial(normalizer, norm="max", axis=axis)
    case _:
      raise ValueError(f"Unknown scaler method {method}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 30, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : losses.py
project  : nadl
license  : GPL-3.0+

Extra Loss functions.
  - focal loss
  - dice loss
"""
from typing import Literal
import jax
import jax.numpy as jnp
from optax import sigmoid_binary_cross_entropy
from jax.nn import log_softmax, log_sigmoid, sigmoid, softmax
from .utils import classit
from .metrics import dice_coef


def sigmoid_focal_loss(
  logits: jax.Array, labels: jax.Array, alpha: float = 0.25, gamma: float = 2.0
) -> jax.Array:
  """Focal loss."""
  p = sigmoid(logits)
  ce_loss = sigmoid_binary_cross_entropy(logits, labels)
  p_t = p * labels + (1 - p) * (1 - labels)
  loss = ce_loss * ((1 - p_t) ** gamma)

  alpha_t = labels * alpha + (1 - labels) * (1 - alpha)
  return alpha_t * loss


def softmax_focal_loss(
  logits: jax.Array, labels: jax.Array, alpha: float = 0.25, gamma: float = 2.0
) -> jax.Array:
  """Focal loss."""
  focus = jnp.power(-jax.nn.softmax(logits, axis=-1) + 1.0, gamma)
  loss = -labels * alpha * focus * jax.nn.log_softmax(logits, axis=-1)
  return jnp.sum(loss, axis=-1)


def dice_loss(
  logits: jax.Array,
  labels: jax.Array,
  method: Literal[None, "sigmoid", "softmax"] = "sigmoid",
) -> jax.Array:
  """Dice loss."""
  if labels.ndim == 1:
    labels = labels[:, None]
  if logits.ndim == 1:
    logits = classit(logits[:, None], method=method)
  return jax.vmap(dice_coef)(labels, logits)
